@@ -78,10 +78,9 @@ struct _XfconfBackendPerchannelXml
 
     gchar *config_save_path;
 
-    GTree *channels;
+    GHashTable *channels;
 
     guint save_id;
-    GSList *dirty_channels;
 
     XfconfPropertyChangedFunc prop_changed_func;
     gpointer prop_changed_data;
@@ -96,6 +95,7 @@ typedef struct
 {
     GNode *properties;
     gboolean locked;
+    gboolean dirty;
 } XfconfChannel;
 
 typedef struct
@@ -176,7 +176,7 @@ static void xfconf_backend_perchannel_xml_register_property_changed_func(XfconfB
                                                                          gpointer user_data);
 
 static void xfconf_backend_perchannel_xml_schedule_save(XfconfBackendPerchannelXml *xbpx,
-                                                        const gchar *channel_name);
+                                                        XfconfChannel *channel);
 
 static XfconfChannel *xfconf_backend_perchannel_xml_create_channel(XfconfBackendPerchannelXml *xbpx,
                                                                    const gchar *channel_name);
@@ -223,10 +223,9 @@ xfconf_backend_perchannel_xml_class_init(XfconfBackendPerchannelXmlClass *klass)
 static void
 xfconf_backend_perchannel_xml_init(XfconfBackendPerchannelXml *instance)
 {
-    instance->channels = g_tree_new_full((GCompareDataFunc)g_ascii_strcasecmp,
-                                         NULL,
-                                         (GDestroyNotify)g_free,
-                                         (GDestroyNotify)xfconf_channel_destroy);
+    instance->channels = g_hash_table_new_full(g_str_hash, g_str_equal,
+                                               (GDestroyNotify)g_free,
+                                                (GDestroyNotify)xfconf_channel_destroy);
 }
 
 static void
@@ -237,12 +236,10 @@ xfconf_backend_perchannel_xml_finalize(GObject *obj)
     if(xbpx->save_id) {
         g_source_remove(xbpx->save_id);
         xbpx->save_id = 0;
+        xfconf_backend_perchannel_xml_flush(XFCONF_BACKEND(xbpx), NULL);
     }
 
-    if(xbpx->dirty_channels)
-        xfconf_backend_perchannel_xml_flush(XFCONF_BACKEND(xbpx), NULL);
-
-    g_tree_destroy(xbpx->channels);
+    g_hash_table_destroy(xbpx->channels);
 
     g_free(xbpx->config_save_path);
 
@@ -297,7 +294,7 @@ xfconf_backend_perchannel_xml_set(XfconfBackend *backend,
                                   GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     XfconfProperty *cur_prop;
 
     if(!channel) {
@@ -309,10 +306,7 @@ xfconf_backend_perchannel_xml_set(XfconfBackend *backend,
 #endif
         if(!channel) {
 #ifdef XFCONF_ENABLE_CHECKS
-            if(error && *error) {
-                g_error_free(*error);
-                *error = NULL;
-            }
+            g_clear_error(error);
 #endif
             channel = xfconf_backend_perchannel_xml_create_channel(xbpx,
                                                                    channel_name);
@@ -352,7 +346,7 @@ xfconf_backend_perchannel_xml_set(XfconfBackend *backend,
             xbpx->prop_changed_func(backend, channel_name, property, xbpx->prop_changed_data);
     }
 
-    xfconf_backend_perchannel_xml_schedule_save(xbpx, channel_name);
+    xfconf_backend_perchannel_xml_schedule_save(xbpx, channel);
 
     return TRUE;
 }
@@ -365,7 +359,7 @@ xfconf_backend_perchannel_xml_get(XfconfBackend *backend,
                                   GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     XfconfProperty *cur_prop;
     GValue *value_to_get = NULL;
 
@@ -454,7 +448,7 @@ xfconf_backend_perchannel_xml_get_all(XfconfBackend *backend,
                                       GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     GNode *props_tree;
     gchar cur_path[MAX_PROP_PATH], *p;
 
@@ -499,7 +493,7 @@ xfconf_backend_perchannel_xml_exists(XfconfBackend *backend,
                                      GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     XfconfProperty *prop;
 
     if(!channel) {
@@ -511,10 +505,7 @@ xfconf_backend_perchannel_xml_exists(XfconfBackend *backend,
 #endif
         if(!channel) {
 #ifdef XFCONF_ENABLE_CHECKS
-            if(error && *error) {
-                g_error_free(*error);
-                *error = NULL;
-            }
+            g_clear_error(error);
 #endif
 
             *exists = FALSE;
@@ -588,19 +579,8 @@ do_reset_channel(XfconfBackend *backend,
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
     gchar *filename;
-    GSList *dirty;
     PropChangeData pdata;
 
-    if((dirty = g_slist_find_custom(xbpx->dirty_channels, channel_name,
-                                    (GCompareFunc)g_ascii_strcasecmp)))
-    {
-        xbpx->dirty_channels = g_slist_remove(xbpx->dirty_channels, dirty);
-        if(!xbpx->dirty_channels && xbpx->save_id) {
-            g_source_remove(xbpx->save_id);
-            xbpx->save_id = 0;
-        }
-    }
-
     pdata.xbpx = xbpx;
     pdata.channel_name = channel_name;
     g_node_traverse(properties, G_POST_ORDER, G_TRAVERSE_ALL, -1,
@@ -609,7 +589,7 @@ do_reset_channel(XfconfBackend *backend,
     /* we could probably prune the existing proptree, or even just leave
      * it as-is, but it's easier to just kill it.  it'll get reloaded later
      * from the system file (if any) if needed. */
-    g_tree_remove(xbpx->channels, channel_name);
+    g_hash_table_remove(xbpx->channels, channel_name);
 
     /* regardless of whether or not we have a system file, we don't need
      * the user file anymore */
@@ -637,7 +617,7 @@ xfconf_backend_perchannel_xml_reset(XfconfBackend *backend,
                                     GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
 
     if(!channel) {
         channel = xfconf_backend_perchannel_xml_load_channel(xbpx, channel_name,
@@ -692,7 +672,7 @@ xfconf_backend_perchannel_xml_reset(XfconfBackend *backend,
         }
     }
 
-    xfconf_backend_perchannel_xml_schedule_save(xbpx, channel_name);
+    xfconf_backend_perchannel_xml_schedule_save(xbpx, channel);
 
     return TRUE;
 }
@@ -729,6 +709,7 @@ xfconf_backend_perchannel_xml_list_channels(XfconfBackend *backend,
 
         g_dir_close(dir);
     }
+    g_strfreev(dirs);
 
     return TRUE;
 }
@@ -741,7 +722,7 @@ xfconf_backend_perchannel_xml_is_property_locked(XfconfBackend *backend,
                                                  GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     XfconfProperty *prop = NULL;
 
     if(!channel) {
@@ -757,15 +738,29 @@ xfconf_backend_perchannel_xml_is_property_locked(XfconfBackend *backend,
     return TRUE;
 }
 
+static void
+xfconf_backend_perchannel_xml_flush_get_dirty(gpointer key,
+                                              gpointer value,
+                                              gpointer user_data)
+{
+    XfconfChannel *channel = value;
+    GSList **dirty = user_data;
+    if(channel->dirty)
+        *dirty = g_slist_prepend(*dirty, key);
+}
+
 static gboolean
 xfconf_backend_perchannel_xml_flush(XfconfBackend *backend,
                                     GError **error)
 {
     XfconfBackendPerchannelXml *xbpx = XFCONF_BACKEND_PERCHANNEL_XML(backend);
-    GSList *l;
+    GSList *dirty = NULL, *l;
 
-    for(l = xbpx->dirty_channels; l; l = l->next)
+    g_hash_table_foreach(xbpx->channels, xfconf_backend_perchannel_xml_flush_get_dirty, &dirty);
+
+    for(l = dirty; l; l = l->next)
         xfconf_backend_perchannel_xml_flush_channel(xbpx, l->data, error);
+    g_slist_free(dirty);
 
     TRACE("exiting, flushed all channels");
 
@@ -796,7 +791,7 @@ xfconf_proptree_lookup_node(GNode *proptree,
 
     g_return_val_if_fail(PROP_NAME_IS_VALID(name), NULL);
 
-    parts = g_strsplit_set(name+1, "/", -1);
+    parts = g_strsplit(name+1, "/", -1);
     parent = proptree;
 
     for(i = 0; parts[i]; ++i) {
@@ -996,24 +991,13 @@ xfconf_backend_perchannel_xml_save_timeout(gpointer data)
 
 static void
 xfconf_backend_perchannel_xml_schedule_save(XfconfBackendPerchannelXml *xbpx,
-                                            const gchar *channel_name)
+                                            XfconfChannel *channel)
 {
-    if(!g_slist_find_custom(xbpx->dirty_channels, channel_name,
-                            (GCompareFunc)g_ascii_strcasecmp))
-    {
-        gpointer orig_key = NULL, val = NULL;
-
-        if(!g_tree_lookup_extended(xbpx->channels, channel_name, &orig_key, &val)) {
-            g_warning("Attempt to schedule save for a nonexistent channel.");
-            return;
-        }
-
-        xbpx->dirty_channels = g_slist_prepend(xbpx->dirty_channels, orig_key);
-    }
-
     if(xbpx->save_id)
         g_source_remove(xbpx->save_id);
 
+    channel->dirty = TRUE;
+
     xbpx->save_id = g_timeout_add_seconds(WRITE_TIMEOUT,
                                           xfconf_backend_perchannel_xml_save_timeout,
                                           xbpx);
@@ -1026,7 +1010,8 @@ xfconf_backend_perchannel_xml_create_channel(XfconfBackendPerchannelXml *xbpx,
     XfconfChannel *channel;
     XfconfProperty *prop;
 
-    if((channel = g_tree_lookup(xbpx->channels, channel_name))) {
+    channel = g_hash_table_lookup(xbpx->channels, channel_name);
+    if(channel) {
         g_warning("Attempt to create channel when one already exists.");
         return channel;
     }
@@ -1035,7 +1020,7 @@ xfconf_backend_perchannel_xml_create_channel(XfconfBackendPerchannelXml *xbpx,
     prop = g_slice_new0(XfconfProperty);
     prop->name = g_strdup("/");
     channel->properties = g_node_new(prop);
-    g_tree_insert(xbpx->channels, g_ascii_strdown(channel_name, -1), channel);
+    g_hash_table_insert(xbpx->channels, g_ascii_strdown(channel_name, -1), channel);
 
     return channel;
 }
@@ -1496,6 +1481,7 @@ xfconf_backend_perchannel_xml_end_elem(GMarkupParseContext *context,
 
         case ELEM_PROPERTY:
             /* FIXME: use stacks here */
+            g_free(state->list_property);
             state->list_property = NULL;
             state->list_value = NULL;
 
@@ -1581,78 +1567,65 @@ xfconf_backend_perchannel_xml_merge_file(XfconfBackendPerchannelXml *xbpx,
                                          GError **error)
 {
     gboolean ret = FALSE;
-    gchar *file_contents = NULL;
-    GMarkupParseContext *context = NULL;
+    GMappedFile *mmap_file;
+    gchar *file_contents;
+    gsize length;
+    GMarkupParseContext *context;
+    XmlParserState *state;
+    GError *error2 = NULL;
     GMarkupParser parser = {
         xfconf_backend_perchannel_xml_start_elem,
         xfconf_backend_perchannel_xml_end_elem,
         /* xfconf_backend_perchannel_xml_text_elem, */
         NULL,
     };
-    XmlParserState state;
-    int fd = -1;
-    struct stat st;
-#ifdef HAVE_MMAP
-    void *addr = NULL;
-#endif
 
     TRACE("entering (%s)", filename);
 
-    memset(&state, 0, sizeof(XmlParserState));
-    state.channel = channel;
-    state.xbpx = xbpx;
-    state.cur_elem = ELEM_NONE;
-    state.is_system_file = is_system_file;
-
-    fd = open(filename, O_RDONLY, 0);
-    if(fd < 0)
-        goto out;
-
-    if(fstat(fd, &st))
-        goto out;
-
-#ifdef HAVE_MMAP
-    addr = mmap(NULL, st.st_size, PROT_READ, MAP_FILE | MAP_SHARED, fd, 0);
-    if(addr != MAP_FAILED)
-        file_contents = addr;
-#endif
-
-    if(!file_contents) {
-        file_contents = g_malloc(st.st_size);
-        if(read(fd, file_contents, st.st_size) != st.st_size)
-            goto out;
+    /* we first try to load a mapped file, if this fails (no mmap
+     * implementation is a possible cause) we fall back to normal file
+     * loading */
+    mmap_file = g_mapped_file_new(filename, FALSE, NULL);
+    if(G_LIKELY(mmap_file != NULL)) {
+        file_contents = g_mapped_file_get_contents(mmap_file);
+        length = g_mapped_file_get_length(mmap_file);
+        DBG("successfully loaded mapped file");
+    } else if(!g_file_get_contents(filename, &file_contents, &length, error)) {
+        return FALSE;
     }
 
-    DBG("got file(size=%lu): %s", st.st_size, file_contents);
+    state = g_slice_new0(XmlParserState);
+    state->channel = channel;
+    state->xbpx = xbpx;
+    state->cur_elem = ELEM_NONE;
+    state->is_system_file = is_system_file;
 
-    context = g_markup_parse_context_new(&parser, 0, &state, NULL);
-    if(!g_markup_parse_context_parse(context, file_contents, st.st_size, error)
-       || !g_markup_parse_context_end_parse(context, error))
-    {
+    DBG("got file(size=%lu): %s", length, file_contents);
+
+    context = g_markup_parse_context_new(&parser, 0, state, NULL);
+    if(g_markup_parse_context_parse(context, file_contents, length, &error2)
+       && g_markup_parse_context_end_parse(context, &error2)) {
+        ret = TRUE;
+    } else {
         g_warning("Error parsing xfconf config file \"%s\": %s", filename,
-                  error && *error ? (*error)->message : "(?)");
-        goto out;
+                  error2 ? error2->message : "(?)");
+        if(error)
+          *error = error2;
+        else
+          g_error_free(error2);
     }
 
-    ret = TRUE;
-
-out:
     TRACE("exiting");
 
+    g_slice_free(XmlParserState, state);
+
     if(context)
         g_markup_parse_context_free(context);
 
-#ifdef HAVE_MMAP
-    if(addr) {
-        munmap(addr, st.st_size);
-        file_contents = NULL;
-    }
-#endif
-
-    g_free(file_contents);
-
-    if(fd >= 0)
-        close(fd);
+    if(mmap_file)
+        g_mapped_file_free(mmap_file);
+    else
+        g_free(file_contents);
 
     return ret;
 }
@@ -1705,7 +1678,7 @@ xfconf_backend_perchannel_xml_load_channel(XfconfBackendPerchannelXml *xbpx,
                                                  channel, NULL);
     }
 
-    g_tree_insert(xbpx->channels, g_ascii_strdown(channel_name, -1), channel);
+    g_hash_table_insert(xbpx->channels, g_ascii_strdown(channel_name, -1), channel);
 
 out:
     g_strfreev(filenames);
@@ -1917,11 +1890,13 @@ xfconf_backend_perchannel_xml_flush_channel(XfconfBackendPerchannelXml *xbpx,
                                             GError **error)
 {
     gboolean ret = FALSE;
-    XfconfChannel *channel = g_tree_lookup(xbpx->channels, channel_name);
+    XfconfChannel *channel = g_hash_table_lookup(xbpx->channels, channel_name);
     GNode *child;
     gchar *filename = NULL, *filename_tmp = NULL;
     FILE *fp = NULL;
 
+    DBG("Flushed dirty channel \"%s\"", channel_name);
+
     if(!channel) {
         if(error) {
             g_set_error(error, XFCONF_ERROR,
@@ -1994,5 +1969,7 @@ out:
     g_free(filename);
     g_free(filename_tmp);
 
+    channel->dirty = FALSE;
+
     return ret;
 }
